#include <pthread.h>
#include <alsa/asoundlib.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <math.h>
#include <stdlib.h>
#include <stdio.h>
#include <sys/types.h>
#include <dirent.h>
#define true 1

/*
Definicoes de formato wave.
entrada PCM com um canal
taxa de gravação de 44100 (equivalente a qualidade de CDs)
  */
#define WAVE_FORMAT 0x001 //PCM
#define WAVE_CHANNELS 1
#define WAVE_SAMPLE_RATE 44100
#define WAVE_BITS_PER_SAMPLE 16
#define WAVE_FRAME_SIZE ((WAVE_BITS_PER_SAMPLE + 7) >> 3 * WAVE_CHANNELS) 
#define WAVE_BYTES_PER_SECOND (WAVE_SAMPLE_RATE * WAVE_FRAME_SIZE) 

#define FRAMES_IN_BLOCK 0x1000
#define BLOCKSIZE (FRAMES_IN_BLOCK * (WAVE_BITS_PER_SAMPLE / 8))
#define PCM_FORMAT SND_PCM_FORMAT_S16_LE

#define N 512
#define QTD_FILTER 20
#define N_MFCC QTD_FILTER 
#define OVERLAY 64

#define LEN_PART_MIN 10
#define LEN_MIN 20



/*frequencias centrais em Hz*/
static int frequency[QTD_FILTER] = {100, 200, 300, 400,
				500, 600, 700, 800, 900, 1000, 1148,
				1318, 1514, 1737, 1995, 2291, 2630, 3020, 3467, 4000};


typedef struct voice_signal
	{
		int16_t *buffer;/*inteiro de 16 bits*/
		unsigned int number;
	} voice_signal;

struct voice_signal_list
{
	voice_signal vs;
	struct voice_signal_list *next;
};



/*
  * Estrutura constituindo um quadro.
  * Os valores no quadro são as intensidades da representação da função no domínio da frequência.
  * Magnitudes (duplo), as intensidades dos valores Fourier transformado.
  */
typedef struct frame
	{
		double magnitudes[N];
	} frame;

typedef struct mfcc_frame
	{
		double features[N_MFCC];
	} mfcc_frame;

typedef struct comp
	{
		double real;
		double imag;
	}comp;


struct block
{
	void *buf;
	struct block *next;
};


struct block *head = NULL;


typedef struct _word
	{
		mfcc_frame *frames;
		unsigned int n;
		char *name;
		struct _word *next;
	} word;


snd_pcm_t *handle = NULL;
int finish = 2;
unsigned int blocknumber = 0;


/*Arquivo atualmente aberto*/
static FILE *f = NULL;
/*cabeçalho WAVE*/
static char header[45] = "RIFF1024WAVEfmt 10241212102410241212data1024";


void capture(void)
{
	printf("funcao-> capture\n");
	struct block *now = NULL;
	int r;
	blocknumber = 0;
	head = malloc(sizeof(struct block));
	head->buf = NULL;
	head->next = NULL;
	now = head;
	while (!finish)
	{
		unsigned int framenumber = FRAMES_IN_BLOCK;
		blocknumber++;
		now->next = malloc(sizeof(struct block));
		now = now->next;
		now->buf = malloc(BLOCKSIZE);
		now->next = NULL;
		do
		{
			r = snd_pcm_readi(handle, now->buf + (FRAMES_IN_BLOCK - framenumber) * (WAVE_BITS_PER_SAMPLE / 8), framenumber);
			if (r < 0)
				fprintf(stderr, "AVISO: Erro ao ler Áudio\n");
			else
				framenumber -= r;
		} while (framenumber > 0);
	}
	finish = 2;
}

/*funcao que captura o audio -> buffer
Verificação dos parametros de hardware
*/
int capture_start(const char *device)
{
	unsigned int freq = WAVE_SAMPLE_RATE;
	int freq_adjust_direction = 1;
	snd_pcm_hw_params_t *hw_params = NULL;
	pthread_t t;

	if (snd_pcm_open(&handle, !device ? "default" : device, SND_PCM_STREAM_CAPTURE, 0) < 0)
	{
		fprintf(stderr, "Nao foi possivel abrir dispositivo %s\n", !device ? "default" : device);
		return -1;
	}

	if (snd_pcm_hw_params_malloc(&hw_params) < 0)
	{
		fprintf(stderr, "Falha ao alocar memória para os parâmetros de hardware\n");
		return -2;
	}

	if (snd_pcm_hw_params_any(handle, hw_params) < 0)
	{
		fprintf(stderr, "Falha ao obter os parâmetros de hardware padrão\n");
		return -3;
	}

	if (snd_pcm_hw_params_set_access(handle, hw_params, SND_PCM_ACCESS_RW_INTERLEAVED) < 0)
	{
		fprintf(stderr, "Definição erro de leitura - \\ acesso de gravação\n");
		return -4;
	}
	if (snd_pcm_hw_params_set_format(handle, hw_params, PCM_FORMAT) < 0)
	{
		fprintf(stderr, "Erro formato de definição PCM\n");
		return -5;
	}
	if (snd_pcm_hw_params_set_rate_near(handle, hw_params, &freq, &freq_adjust_direction) < 0)
	{
		fprintf(stderr, "Falha ao definir a freqüência\n");
		return -6;
	}
	if (snd_pcm_hw_params_set_channels(handle, hw_params, WAVE_CHANNELS) < 0)
	{
		fprintf(stderr, "Falha ao definir o número de canais\n");
		return -7;
	}

	if (snd_pcm_hw_params(handle, hw_params) < 0)
	{
		fprintf(stderr, "Erro ao definir os parâmetros de hardware\n");
		return -8;
	}

	snd_pcm_hw_params_free(hw_params);

	if (snd_pcm_prepare(handle) < 0)
	{
		fprintf(stderr, "Erro no handle\n");
		return -9;
	}

	finish = 0;
	/*Pthread para iniciar a gravação
	A função capture() iniciara em 
	trhead paralela a da execução principal

	*/
	if (pthread_create(&t, NULL, (void *)capture, NULL) < 0)
	{
		fprintf(stderr, "Falha ao iniciar tópicos de gravação\n");
		return -10;
	}
	return 0;
}

void capture_stop(unsigned int *size, void **buffer)
{
	printf("funcao-> capture_stop \n");
	unsigned int i = 0;
	struct block *now = NULL;
	void **adresses = NULL;

	finish = 1;
	while (finish != 2);

	snd_pcm_close(handle);
	*buffer = malloc(BLOCKSIZE * blocknumber);
	*size = BLOCKSIZE * blocknumber;
       	adresses = malloc(blocknumber * sizeof(void *));

	now = head;
	for (i = 0; i < blocknumber; i++)
	{
		now = now->next;
		adresses[i] = now;
		memcpy(*buffer + (i * BLOCKSIZE), now->buf, BLOCKSIZE);
		free(now->buf);
	}
	for (i = 0; i < blocknumber; i++)
		free(adresses[i]);
	free(head);
	free(adresses);
}

/*funcao para gravacao de audio -> disco*/
int record(const char *device, const void *buffer, const unsigned int size)
{
	printf("funcao-> record\n");
	int r = 0;
	unsigned int freq = WAVE_SAMPLE_RATE;
	unsigned int framenumber = size / (WAVE_BITS_PER_SAMPLE / 8);
	snd_pcm_hw_params_t *hw_params = NULL;

	if (snd_pcm_open(&handle, !device ? "default" : device, SND_PCM_STREAM_PLAYBACK, 0) < 0)
	{
		fprintf(stderr, "Dispositivo não foi possível abrir %s \n", !device ? "default" : device);
		return -1;
	}

	if (snd_pcm_hw_params_malloc(&hw_params) < 0)
	{
		fprintf(stderr, "Falha ao alocar memória para os parâmetros de hardware\n");
		return -2;
	}

	if (snd_pcm_hw_params_any(handle, hw_params) < 0)
	{
		fprintf(stderr, "Falha ao obter os parâmetros de hardware padrão\n");
		return -3;
	}

	if (snd_pcm_hw_params_set_access(handle, hw_params, SND_PCM_ACCESS_RW_INTERLEAVED) < 0)
	{
		fprintf(stderr, "Definição erro de leitura - \\ acesso de gravação\n");
		return -4;
	}
	if (snd_pcm_hw_params_set_format(handle, hw_params, PCM_FORMAT) < 0)
	{
		fprintf(stderr, "Erro na definicao de formato PCM\n");
		return -5;
	}
	if (snd_pcm_hw_params_set_rate_near(handle, hw_params, &freq, 0) < 0)
	{
		fprintf(stderr, "Erra na definicao da frequencia\n");
		return -6;
	}
	if (snd_pcm_hw_params_set_channels(handle, hw_params, WAVE_CHANNELS) < 0)
	{
		fprintf(stderr, "Falha ao definir o número de canais\n");
		return -7;
	}

	if (snd_pcm_hw_params(handle, hw_params) < 0)
	{
		fprintf(stderr, "Falha ao definir parametros de hardware\n");
		return -8;
	}

	snd_pcm_hw_params_free(hw_params);

	if (snd_pcm_prepare(handle) < 0)
	{
		fprintf(stderr, "Erro preparando Handles\n");
		return -9;
	}

	do
	{
		r = snd_pcm_writei(handle, buffer + (size - framenumber * (WAVE_BITS_PER_SAMPLE / 8)), framenumber);
		if (r < 0)
			fprintf(stderr, "AVISO: Falha ao gravar Áudio\n");
		else
			framenumber -= r;
	} while (framenumber > 0);

	snd_pcm_close(handle);
	return 0;
}


/*
Prepara o cabeçalho wave
  * Preenche o cabeçalho comparação com os valores válidos, que são especificados 
em wave.h e o arquivo e bloco de tamanho correto, que é derivado dos parâmetros.
  * (Const unsigned int) tamanho do arquivo
*/
void prepare_header(const unsigned int size)
{
	printf("funcao-> prepare_header\n");
	int values[9] = {size - 8, 16, WAVE_FORMAT, WAVE_CHANNELS, WAVE_SAMPLE_RATE, WAVE_BYTES_PER_SECOND, WAVE_FRAME_SIZE, WAVE_BITS_PER_SAMPLE, size - 44};

	memcpy(header + 4, values, 4);
	memcpy(header + 16, values + 1, 4);
	memcpy(header + 20, values + 2, 2);
	memcpy(header + 22, values + 3, 2);
	memcpy(header + 24, values + 4, 4);
	memcpy(header + 28, values + 5, 4);
	memcpy(header + 32, values + 6, 2);
	memcpy(header + 34, values + 7, 2);
	memcpy(header + 40, values + 8, 4);

}

int write_pcm(const void *buffer, const unsigned int size, const char *path)
{
	FILE *file = NULL;
	printf("funcao-> write_pcm\n");

	if (path != NULL)
		file = fopen(path, "w");
	else
	{
		rewind(f);
		file = f;
	}
	if (file == NULL)
		return -1;

	prepare_header(size + 44);

	fwrite(header, 44, 1, file);
	fwrite(buffer, size, 1, file);

	if (path != NULL)
		fclose(file);
	return 0;
}



int split(int16_t *buffer, unsigned int n, voice_signal **signals)
{
	unsigned int i = 0, j = 0, k = 0, start_pos = 0, len_sig = 0, len_nothing = 0, count = 0;
	char started = 0;
	double means[n / N], sds[n / N];
	double sdmean = 0;
	void **free_list = NULL;
	struct voice_signal_list *head = malloc(sizeof(struct voice_signal_list));
	struct voice_signal_list *now = head;

	puts("split");
	head->next = NULL;

	// Calcula os desvios padrão de cada quadro
	for (i = 0; i < n / N; i++)
	{
		means[i] = 0;
		for (j = 0; j < N; j++)
			means[i] += ((int16_t *)buffer)[i * N + j];
		means[i] /= N;
		sds[i] = 0;
		for (j = 0; j < N; j++)
			sds[i] += pow(((int16_t *)buffer)[i * N +j] - means[i], 2);
		sds[i] = sqrt(sds[i]) / (N - 1);
		sdmean += sds[i];
	}
	sdmean /= (n / N);

	// comparação dos desvios-padrão do quadro atual com o desvio padrão médio, para descobrir onde os sinais ocorrem
	for (i = 0; i < n / N; i++)
	{
		//sinal disponivel
		if (sdmean < sds[i])
		{
			//inicializacao sinal start
			if (!started)
			{
				started = 1;
				start_pos = i;
				len_nothing = 0;
			}
			len_sig++;
		}
		//sinal muito fraco
		else
			// (sinal - curto silêncio - sinal) deve ser interpretado como um único sinal
			if (len_sig > LEN_PART_MIN)
			{
				len_nothing = 0;
				len_sig = 0;
			}
			else
			{
				// Duração do silêncio
				len_nothing++;
				if (started && i - LEN_MIN > start_pos && (len_nothing > 2 * LEN_MIN || i == n / N - 1))
				{
					 // Cria uma nova entrada 
					now->next = malloc(sizeof(struct voice_signal_list));
					now = now->next;
					now->next = NULL;
					now->vs.number = (i - start_pos) * N;
					now->vs.buffer = malloc(sizeof(int16_t) * now->vs.number);

					for (j = start_pos; j < i; j++)
					{
						for (k = 0; k < N; k++)
							now->vs.buffer[(j - start_pos) * N + k] = ((int16_t *)buffer)[j * N + k];
					}

					count++;
					started = 0;
				}
			}
	}

	//Preenche o sinal parâmetro com os sinais de voz reconhecidos
	*signals = malloc(sizeof(voice_signal) * count);
	free_list = malloc(sizeof(void *) * count);
	now = head;
	for (i = 0; i < count; i++)
	{
		now = now->next;
		free_list[i] = now;
		(*signals)[i] = now->vs;
	}
	for (i = 0; i < count; i++)
		free(free_list[i]);
	free(free_list);
	free(head);

	return count;
}


void fft(comp *time, comp *freq, unsigned int n)
{
	double pi = 4 * atan(1.0);
	unsigned int i = 0, j = 0, k = 0;
	unsigned int bits = 0;
	puts("fft");
	if (n == 1)
	{
		freq[0] = time[0];
		return;
	}

	/*bits necessários para a representação binária de n*/
	for (i = n; i > 1; i /= 2)
		bits++;

	for (i = 0; i < n; i++)
	{
		int reversed = 0;
		for (j = 0; j < bits; j++)
			reversed |= (((i >> j) % 2) << (bits - j - 1));
		freq[reversed] = time[i];
	}

	for (i = bits; i > 0; i--)
	{
		unsigned int length_groups = (n >> (i - 1));
		unsigned int num_groups = n / length_groups;
		for (j = 0; j < num_groups; j++)
		{
			for (k = 0; k < length_groups / 2; k++)
			{
				unsigned int off1 = length_groups * j + k;
				unsigned int off2 = off1 + length_groups / 2;
				comp off1_val, off2_val;
				double cosine = cos((-2 * pi * k) / length_groups);
				double sine = sin((-2 * pi * k) / length_groups);
				off1_val.real = freq[off1].real + (freq[off2].real * cosine - freq[off2].imag * sine);
				off1_val.imag = freq[off1].imag + (freq[off2].imag * cosine + freq[off2].real * sine);
				off2_val.real = freq[off1].real - (freq[off2].real * cosine - freq[off2].imag * sine);
				off2_val.imag = freq[off1].imag - (freq[off2].imag * cosine + freq[off2].real * sine);
				freq[off1] = off1_val;
				freq[off2] = off2_val;
			}
		}
	}
	for (i = 0; i < n; i++)
	{
		freq[i].real /= n;
		freq[i].imag /= n;
		freq[i].imag *= -1;
	}
}

unsigned int make_frames_hamming(int16_t *buffer, unsigned int n, frame **frames)
{
	unsigned int id_frame = (n / (N - OVERLAY)) - 1;
	comp *result = malloc(sizeof(comp) * id_frame * N);
	comp *data = malloc(sizeof(comp) * id_frame * N);
	unsigned int pos = 0, i = 0, j = 0;
	double pi = 4 * atan(1.0);
	puts("make_frames_hamming");
	printf("n frames-> %d\n", n);
	*frames = malloc(sizeof(frame) * id_frame);
	for (i = 0; i < id_frame; i++)
	{
		pos = (i + 1) * (N - 64);
		for (j = 0; j < N; j++)
		{
			data[i * N + j].real = buffer[pos + j] * (0.54 + 0.46 * cos(2 * pi * (j - N / 2) / N));
			data[i * N + j].imag = 0;
		}
	}
	for (i = 0; i < id_frame; i++)
		fft(data + i * N, result + i * N, N);

	for (i = 0; i < id_frame; i++)
		for (j = 0; j < N; j++)
			(*frames)[i].magnitudes[j] = sqrt(pow(result[i * N + j].real, 2) + pow(result[i * N + j].imag, 2));

	for (i = 0; i < id_frame; i++)
	{
		double sig = 0;
		for (j = 0; j < N; j++)
			sig += (*frames)[i].magnitudes[j];
		sig /= N;
		for (j = 0; j < N; j++)
			(*frames)[i].magnitudes[j] /= sig;
	}
	free(data);
	free(result);
	return id_frame;
}


void mfcc_features(frame *frames, unsigned int n, mfcc_frame *mfcc_frames)
{
	unsigned int i = 0, j = 0, k = 0;
	double out_filter[QTD_FILTER];
	double filterSpectrum[QTD_FILTER][N] = {{0}};
	double c0 = sqrt(1.0 / QTD_FILTER);
	double cn = sqrt(2.0 / QTD_FILTER);
	double pi = 4 * atan(1.0);

	puts("mfcc_features");
	for (i = 0; i < QTD_FILTER; i++)
	{
		double maxFreq = 0, minFreq = 0, centerFreq = 0;

		if (i == 0)
			minFreq = frequency[0] - (frequency[1] - frequency[0]);
		else
			minFreq = frequency[i - 1];
		centerFreq = frequency[i];

		if (i == QTD_FILTER - 1)
			maxFreq = frequency[QTD_FILTER - 1] + (frequency[QTD_FILTER - 1] - frequency[QTD_FILTER - 2]);
		else
			maxFreq = frequency[i + 1];

		for (j = 0; j < N; j++)
		{
			double freq = 1.0 * j * WAVE_SAMPLE_RATE / N;
			if (freq > minFreq && freq < maxFreq)
				if (freq < centerFreq)
					filterSpectrum[i][j] = 1.0 * (freq - minFreq) / (centerFreq - minFreq);
				else
					filterSpectrum[i][j] = 1 - 1.0 * (freq - centerFreq) / (maxFreq - centerFreq);
			else
				filterSpectrum[i][j] = 0;
		}
	}

	/*Calcula os coeficientes MFCC para cada quadro.*/
	for (i = 0; i < n; i++)
	{
		for (j = 0; j < QTD_FILTER; j++)
		{
			out_filter[j] = 0;
			for (k = 0; k < N; k++)
				out_filter[j] += frames[i].magnitudes[k] * filterSpectrum[j][k];
		}
		for (j = 0; j < N_MFCC; j++)
		{
			mfcc_frames[i].features[j] = 0;

			for (k = 0; k < QTD_FILTER; k++)
				mfcc_frames[i].features[j] += log(fabs(out_filter[k]) + 1e-10) * cos((pi * (2 * k + 1) * j) / (2 * QTD_FILTER));

			if (j)
				mfcc_frames[i].features[j] *= cn;
			else
				mfcc_frames[i].features[j] *= c0;
		}
	}
}


int get_list(word *head)
{
	DIR *d = opendir("palavras");
	FILE *f = NULL;
	struct dirent *dentry = readdir(d);
	unsigned int i = 0, j = 0;
	unsigned int count = 0;
	puts("get_list");

	chdir("palavras");
	while (dentry != NULL)
	{
		if (!memcmp(dentry->d_name, ".", 1))
		{
			dentry = readdir(d);
			continue;
		}

		f = fopen(dentry->d_name, "r");
		head->name = malloc(strlen(dentry->d_name) + 1);
		memcpy(head->name, dentry->d_name, strlen(dentry->d_name) + 1);
		fscanf(f, "%u", &(head->n));
		head->frames = malloc(sizeof(mfcc_frame) * head->n);
		for (i = 0; i < head->n; i++)
			for (j = 0; j < N_MFCC; j++)
				fscanf(f, "%lf", &(head->frames[i].features[j]));
		fclose(f);

		dentry = readdir(d);
		count++;

		head->next = NULL;
		if (dentry != NULL && memcmp(dentry->d_name, ".", 2) && memcmp(dentry->d_name, "..", 3))
		{
			head->next = malloc(sizeof(word));
			head = head->next;
		}
	}
	chdir("..");

	return count;
}


/*Busca de padrão com metodo deterministico, calculada a distancia euclidiana entre o 
padrão buscado e os padrões de referencia*/
double compare(mfcc_frame *mfcc_frames1, unsigned int n1, mfcc_frame *mfcc_frames2, unsigned int n2)
{
	double distances[n1 + 1][n2 + 1];
	double prev_min;
	unsigned int i = 0, j = 0, k = 0;
	
	/*Calcule as distâncias de todos os pares de coeficientes MFCC*/
	for (i = 0; i < n1; i++)
	{
		
		for (j = 0; j < n2; j++)
		{
			distances[i + 1][j + 1] = 0;/*Distancias euclidianas*/
			for (k = 0; k < N_MFCC; k++)
			{
				distances[i + 1][j + 1] += pow(mfcc_frames1[i].features[k] - mfcc_frames2[j].features[k], 2);
			
			}
			distances[i + 1][j + 1] = sqrt(distances[i + 1][j + 1]);
		}
	}

	/* Preenchendo as fronteiras com o infinito*/
	for (i = 0; i <= n1; i++)
		distances[i][0] = atof("Inf");
	for (i = 0; i <= n2; i++)
		distances[0][i] = atof("Inf");
	distances[0][0] = 0;
	
	for (i = 1; i <= n1; i++)
		for (j = 1; j <= n2; j++)
		{
			prev_min = distances[i - 1][j];
			if (distances[i - 1][j - 1] < prev_min)
				prev_min = distances[i - 1][j - 1];
			if (distances[i][j - 1] < prev_min)
				prev_min = distances[i][j - 1];
			
			distances[i][j] += prev_min;
		}

	/*Normalizar a distância*/
	return distances[n1][n2] / sqrt(pow(n1, 2) + pow(n2, 2));
}





void get_word(int flag)
{
	unsigned int size = 0, n = 0, i = 0;
	char ans[500];
	int frame_n;
	void *buffer = NULL;
	frame *frames = NULL;
	mfcc_frame *mfcc_frames = NULL;
	voice_signal *signals = NULL;
	FILE *out;
	printf("\nFale agora e pressione enter\n");
	capture_start(NULL);
	sleep(5);
	capture_stop(&size, &buffer);
	n = split(buffer, size / 2, &signals);
	for (i = 0; i < n; i++)
	{
		frame_n = make_frames_hamming(signals[i].buffer, signals[i].number, &frames);
		mfcc_frames = malloc(sizeof(mfcc_frame) * frame_n);
		mfcc_features(frames, frame_n, mfcc_frames);
		if (!flag)
		{
			puts("not new");
		}
		else
		{
			/*laço de reconhecimento*/	
			puts("analisando dados...");
			word *words = malloc(sizeof(word));
			double best = 1e10;
			char *name = NULL;
			void **word_adresses;
			unsigned int n = 0, i = 0, count = 0;
			if ((n = get_list(words)))
			{
				word_adresses = malloc(n * sizeof(void *));
				while (words != NULL)
				{
					double now = compare(mfcc_frames, frame_n, words->frames, words->n);
					word_adresses[count++] = words;
					if (now < best)
					{
						best = now;
						name = words->name;
					}
					words = words->next;
				}
				for (i = 0; i < count; i++)
					free(word_adresses[i]);
				free(word_adresses);
			}
			out = fopen("out.txt", "a+");
			if (best < 3.5)
			{
				fprintf(out, " %s\n", name);
			}
		}
		free(mfcc_frames);
		free(frames);
	}
	if (flag)
		printf("\n");
	free(buffer);
	
}




int main(){
	
	get_word(1);	
	
	return 0;

}




