from hmm import HMM
import numpy as np

def viterbi(hmm, initial_dist, emissions):
    probs = hmm.emission_dist(emissions[0]) * initial_dist
    stack = []

    for emission in emissions[1:]:
        trans_probs = hmm.transition_probs * np.row_stack(probs)
        max_col_ixs = np.argmax(trans_probs, axis=0)
        probs = hmm.emission_dist(emission) * trans_probs[max_col_ixs, np.arange(hmm.num_states)]

        stack.append(max_col_ixs)

    state_seq = [np.argmax(probs)]

    while stack:
        max_col_ixs = stack.pop()
        state_seq.append(max_col_ixs[state_seq[-1]])

    state_seq.reverse()

    return state_seq

def main():
    wiki_emission_probs = np.array([[0.5, 0.6, 0.05, 0.3], 
                                    [0.32, 0.25, 0.35, 0.3]])

    wiki_initial_dist = np.array([[0.6, 0.4]])
    wiki_emissions = [0, 1, 2, 2, 3, 0, 0]

    wiki_transition_probs = np.array([[0.7, 0.3], 
                                    [0.6, 0.4]])

    wiki_hmm = HMM(wiki_transition_probs, wiki_emission_probs)
    print(viterbi(wiki_hmm, wiki_initial_dist, wiki_emissions))

if __name__ == "__main__":
    main()
    
